import typing as tp

import numpy as np


def chunkify(signal: np.ndarray, chunk_size: int = 256, n_overlap: int = 32) -> np.ndarray:
    """todo: see
        - https://numpy.org/doc/stable/reference/generated/numpy.lib.stride_tricks.sliding_window_view.html
        - https://github.com/pydata/bottleneck
    """
    step = chunk_size - n_overlap
    shape = ((signal.shape[-1] - n_overlap) // step, chunk_size)
    strides = (step * signal.strides[-1], signal.strides[-1])
    result = np.lib.stride_tricks.as_strided(
        signal, shape=shape, strides=strides,
    )
    return result


def make_windows(a: np.ndarray, sizes: tp.Tuple[int, ...]) -> tp.List[np.ndarray]:
    """TODO: just create window for max size and then use slices [:, -size:] to make smaller windows
        - but it can be slower because slicing returns a copy instead of view
        - or not: https://scipy-cookbook.readthedocs.io/items/ViewsVsCopies.html#Slice-views
    """
    if any(size < 1 for size in sizes):
        raise ValueError('window size should be >= 1')
    windows = []

    # # multiple sliding_window_view
    # max_size = max(sizes)
    # for w_size in sizes:
    #     windows.append(
    #         np.lib.stride_tricks.sliding_window_view(
    #             a, window_shape=w_size,
    #         )[max_size - w_size:],
    #     )

    # single sliding_window_view and slices
    _sizes = sorted(sizes)
    max_window = np.lib.stride_tricks.sliding_window_view(
        a, window_shape=sizes[-1],
    )
    for w_size in _sizes:
        windows.append(max_window[:, -w_size:])

    return windows



import typing as tp

import numpy as np


def _within_bounds(
    value: tp.Union[float, np.ndarray],
    bounds: tp.Tuple[float, float],
) -> bool:
    a, b = bounds
    if a == b:
        return value == a
    min_, max_ = min(a, b), max(a, b)
    if isinstance(value, np.ndarray):
        return bool(np.all((min_ <= value) & (value <= max_)))
    return min_ <= value <= max_


def minmax_scaler(
    value: tp.Union[float, np.ndarray],
    oldmin: float,
    oldmax: float,
    newmin: float = 0.0,
    newmax: float = 1.0,
) -> tp.Union[float, np.ndarray]:

    if not _within_bounds(value, (oldmin, oldmax)):
        raise ValueError('value should be oldmin <= value <= oldmax')

    if oldmin == oldmax:
        if newmin != newmax:
            raise ValueError('oldmin == oldmax, so newmin == newmax must be true')
        return newmin

    old_settings = np.seterr(over='raise')
    out = (value - oldmin) * (newmax - newmin) / (oldmax - oldmin) + newmin
    np.seterr(**old_settings)
    return out



import typing as tp

import numpy as np


class CircularBuffer:
    """
    actually grows "row by row" instead of collection.deque-like fashion
    """

    def __init__(
        self,
        size: tp.Optional[int] = None,
        dtype: type = np.float32,
        index: int = 0,
        buffer: tp.Optional[np.ndarray] = None,
    ):
        self.dtype = dtype
        self.index = index
        if buffer is not None:
            self.buffer = buffer
            if size is not None:
                raise ValueError('size must be None if buffer is passed')
            self.size = buffer.shape[0]
        else:
            if size is None:
                raise ValueError('size must be passed if buffer is None')
            self.buffer = np.zeros(size, dtype=dtype)
            self.size = size

    def append(self, v: float) -> None:
        self.buffer[self.index] = v
        self.index = (self.index + 1) % self.size

    def extend(self, x: np.ndarray) -> None:
        l = x.shape[0]
        if l <= self.size:
            if self.index + l < self.size:
                self.buffer[self.index:self.index + l] = x
            else:
                n_append_end = self.size - self.index
                self.buffer[self.index:] = x[:n_append_end]
                self.buffer[:l - n_append_end] = x[n_append_end:]
        else:
            mod = l % self.size
            q = mod - (self.size - self.index)
            self.buffer[:q] = x[-q:]
            self.buffer[q:] = x[-self.size:-q]  # pylint: disable=invalid-unary-operand-type
        self.index = (self.index + l) % self.size

    def most_recent(self, n: int) -> np.ndarray:
        if n > self.size:
            raise ValueError(f'n cant be greater than size {n} > {self.size}')
        if n <= self.index:
            return self.buffer[self.index - n:self.index]
        return np.hstack((self.buffer[-(n - self.index):], self.buffer[:self.index]))
